/*
 *  TU Eindhoven
 *  Eindhoven, The Netherlands
 *
 *  Name            :   statespace.h
 *
 *  Author          :   Sander Stuijk (sander@ics.ele.tue.nl)
 *
 *  Date            :   April 5, 2006
 *
 *  Function        :   State-space based analysis techniques
 *
 *  History         :
 *      05-04-06    :   Initial version.
 *
 * $Id: statespace.h,v 1.3 2007/11/13 17:17:55 sander Exp $
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * 
 * In other words, you are welcome to use, share and improve this program.
 * You are forbidden to forbid anyone else to use, share and improve
 * what you give them.   Happy coding!
 */

#ifndef SDF_ANALYSIS_STATESPACE_STATESPACE_H_INCLUDED
#define SDF_ANALYSIS_STATESPACE_STATESPACE_H_INCLUDED

#include "../../base/timed/graph.h"
#include "../../base/algo/components.h"
#include "../../base/algo/repetition_vector.h"
#include "../../resource_allocation/mpsoc_arch/graph.h"

typedef unsigned short TCnt;
typedef unsigned long TBufSize;
typedef double TDtime;

#define TDTIME_MAX      INT_MAX
#define TBUFSIZE_MAX    ULONG_MAX

/**
 * Throughput analysis
 * Compute the throughput of an SDF graph for unconstrained buffer sizes and
 * using auto-concurrency using a state-space traversal.
 */
#include "throughput.h"

/**
 * Binding-aware throughput analysis
 * Computes the throughput of an SDFG mapped to an architecture platform. It
 * is assumed that the platform uses a TDMA resource arbitration mechanism on
 * the processors. An Actor is scheduled on a processor if its input tokens are
 * available and the static-order schedule indicates that the actor is allowed
 * to fire. Actors which are not mapped to a processor, only wait for their
 * input tokens.
 */
#include "tdma_schedule.h"

/**
 * Throughput / storage-space trade-off exploration
 * Analyze the trade-offs between storage distributions and throughput (using
 * auto-concurrency). The search ends as soon as the throughput bound (thrBound)
 * is reached. To find the complete pareto-space, the throughput bound should
 * be set to DOUBLE_MAX.
 */
#include "buffer.h"

/**
 * Throughput / storage-space trade-off exploration
 * Analyze the trade-offs between storage distributions and throughput (using
 * auto-concurrency) under the given binding constraints of the application
 * to the architecture graph. The vector 'bufferChannels' indicates for each
 * channel wether initial tokens can be added to it. This is only true for
 * channels modeling buffer space. When the 'useBoundsOnBufferChannels' flag
 * is true, the number of initial tokens (i.e. storage space) is not enlarged
 * above the current number of initial tokens on the edge (i.e. current storage
 * space). This storage space allocation is then a constraint on the maximum
 * amount of storage space that can be allocated to the channel. The exploration
 * ends as soon as the throughput reaches the throughput bound (thrBound) or all
 * minimal storage distributions are found (default).
 */
#include "bounded_buffer.h"

/**
 * Trace interconnect communication
 * Generates a trace of all communication over the interconnect
 * of a mapped and scheduled SDFG on a NoC-based platform. The relation 
 * between the communication in the transient and periodic part of the
 * schedule is also included.
 */
#include "comm_trace.h"

/**
 * Deadlock analysis
 * Check that an SDFG is deadlock free (i.e. executing each actor as often as
 * indicated by the repetition vector returns the graph to its initial state).
 */
#include "deadlock.h"

//------------------------------------------------------------------------------



/**
 * stateSpaceAbstractDepGraph ()
 * Compute the throughput and abstract dependency graph of an SDFG for
 * unconstrained buffer sizes and using auto-concurrency using a state-space
 * traversal.
 *
 * The abstract dependency graph is stored in the matrices 'nodes' (2D) and
 * 'edges' (3D). The value stored at 'nodes[a][b]' indicates whether there is at
 * least one edge which goes directlt from node a to node b. The value stored at
 * 'edges[a][b][c]' indicates wether the channel c in the SDFG has a depedency
 * edge in the abstract dependency graph from node a to node b.
 */
double stateSpaceAbstractDepGraph(TimedSDFgraph *gr, bool ***nodes, 
        bool ****edges, unsigned long long stackSz = 1000, 
        unsigned long long hashSz = 10000);

/**
 * stateSpaceLatencyAnalysisSelfTimed ()
 * Compute the latency of an SDF graph for unconstrained buffer sizes and
 * using auto-concurrency using a state-space traversal and self-timed 
 * execution.
 */
double stateSpaceLatencyAnalysisSelfTimed(TimedSDFgraph *gr,
        SDFactor *srcActor, SDFactor *dstActor, double *thr = NULL,
        unsigned long long stackSz = 1000, unsigned long long hashSz = 10000);

/**
 * stateSpaceLatencyAnalysisMinimal ()
 * Compute the latency of an SDF graph for unconstrained buffer sizes and
 * using auto-concurrency using a state-space traversal and self-timed 
 * execution.
 */
double stateSpaceLatencyAnalysisMinimal(TimedSDFgraph *gr,
        SDFactor *srcActor, SDFactor *dstActor, double *thr = NULL,
        unsigned long long stackSz = 1000, unsigned long long hashSz = 10000);

/**
 * stateSpaceLatencyAnalysisSingleProc ()
 * Compute the latency of an SDF graph for unconstrained buffer sizes and
 * using auto-concurrency using a state-space traversal and self-timed 
 * execution.
 */
double stateSpaceLatencyAnalysisSingleProc(TimedSDFgraph *gr,
        SDFactor *srcActor, SDFactor *dstActor,
        unsigned long long stackSz = 1000, unsigned long long hashSz = 10000);

/**
 * stateSpaceMinimalLatencySelfTimed ()
 * Compute the minimal latency and maximal throughput of an SDF graph for
 * unconstrained buffer sizes and using auto-concurrency using a state-space
 * traversal and self-timed execution. The source actor is not allowed to
 * execute in a self-timed manner. Its execution is constraint to minimize the
 * latency.
 */
double stateSpaceMinimalLatencySelfTimed(TimedSDFgraph *gr,
        SDFactor *srcActor, SDFactor *dstActor, double *thr,
        unsigned long long stackSz = 1000, unsigned long long hashSz = 10000);

/**
 * stateSpaceBufferAnalysisCapacityConstrainedModel ()
 * Analyze the trade-offs between storage distributions and throughput (using
 * auto-concurrency). Storage space allocations are determined for all channels
 * modelling storage space in the capacity constrained model.
 */
StorageDistributionSet *stateSpaceBufferAnalysisCapacityConstrainedModel(
        TimedSDFgraph *gr, double maxThr);

#endif
