/*
 *  TU Eindhoven
 *  Eindhoven, The Netherlands
 *
 *  Name            :   sdf.cc
 *
 *  Author          :   Sander Stuijk (sander@ics.ele.tue.nl)
 *
 *  Date            :   April 3, 2006
 *
 *  Function        :   Model binding decisions in timed SDFG.
 *
 *  History         :
 *      03-04-06    :   Initial version.
 *
 * $Id: sdf.cc,v 1.2 2008/01/24 12:43:56 sander Exp $
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * 
 * In other words, you are welcome to use, share and improve this program.
 * You are forbidden to forbid anyone else to use, share and improve
 * what you give them.   Happy coding!
 */

#include "sdf.h"

/**
 * extractActorBinding ()
 * The function constructs a vector of pointers to tiles. The pointer at 
 * actorBinding[i] points to the tile the actor with id 'i' is bound to.
 * It is NULL if the actor is not bound.
 */
void extractActorBinding(TimedSDFgraph *g, PlatformGraph *ag,
        Tiles &actorBinding)
{
    // Initialize binding
    actorBinding.resize(g->nrActors());
    for (uint i = 0; i < g->nrActors(); i++)
        actorBinding[i] = NULL;

    // Iterate over the processors in all tiles to find bindings
    for (TilesIter iter = ag->tilesBegin(); iter != ag->tilesEnd(); iter++)
    {
        Tile *t = *iter;
        Processor *p = t->getProcessor();

        // No processor inside this tile?
        if (p == NULL)
            continue;

        // Iterate over the component bindings
        for (ComponentBindingsIter iterB = p->getActorBindings()->begin();
                iterB != p->getActorBindings()->end(); iterB++)
        {
            ComponentBinding *b = *iterB;
            TimedSDFactor *a = (TimedSDFactor*)(b->getComponent());

            // Actor a already mapped?
            if (actorBinding[a->getId()] != NULL)
            {
                EXCEPTION("Actor '%s' mapped to multiple tiles.",
                            a->getName().c_str());
            }
            
            actorBinding[a->getId()] = t;
        }
    }
} 

/**
 * extractChannelBinding ()
 * The function constructs a vector of pointers to connections. The pointer at 
 * channelBinding[i] points to the connection the channel with id 'i' is
 * bound to. It is NULL if the connection is not bound.
 */
void extractChannelBinding(TimedSDFgraph *g, PlatformGraph *ag,
        Connections &channelBinding)
{
    // Initialize binding
    channelBinding.resize(g->nrChannels());
    for (uint i = 0; i < g->nrChannels(); i++)
        channelBinding[i] = NULL;

    // Iterate over the connections in all tiles to find bindings
    for (ConnectionsIter iter = ag->connectionsBegin(); 
            iter != ag->connectionsEnd(); iter++)
    {
        Connection *c = *iter;

        // Iterate over the component bindings
        for (ComponentBindingsIter iterB = c->getChannelBindings()->begin();
                iterB != c->getChannelBindings()->end(); iterB++)
        {
            ComponentBinding *b = *iterB;
            TimedSDFchannel *ch = (TimedSDFchannel*)(b->getComponent());

            // Channel ch already mapped?
            if (channelBinding[ch->getId()] != NULL)
                throw CException("Channel mapped to multiple connections.");
            
            channelBinding[ch->getId()] = c;
        }
    }
}

/**
 * createMappedActor ()
 * The function models an actor mapped to a tile.
 */
void createMappedActor(TimedSDFactor *a, Tile *t)
{
    SDFchannel *c;
    TimedSDFactor::Processor wcrt;
    
    // Set the execution time of the actor as its worst-case response time
    wcrt.type = "wcrt";
    wcrt.execTime = a->getExecutionTime(t->getProcessor()->getType());
    wcrt.stateSize = 0;
    a->addProcessor(&wcrt);
    a->setDefaultProcessor("wcrt");

    // Add self-loop with one token
    c = a->getGraph()->createChannel(a, 1, a, 1, 1);
}

/**
 * createMappedChannelToTile ()
 * The function models a channel mapped to a tile.
 */
void createMappedChannelToTile(TimedSDFchannel *c, Tile *t)
{
    SDFgraph *g;
    TimedSDFchannel *cMem;
        
    // Add channel from dst to src actor to model memory space
    g = c->getGraph();
    cMem = (TimedSDFchannel*) g->createChannel(c->getDstActor(),
                                               c->getDstPort()->getRate(),
                                               c->getSrcActor(),
                                               c->getSrcPort()->getRate(), 0);
    // Initial tokens reflect storage space
    if ((CSize)c->getBufferSize().mem < c->getInitialTokens())
        throw CException("Insufficient memory space to store initial tokens.");
    cMem->setInitialTokens(c->getBufferSize().mem - c->getInitialTokens());
    
    // Channel cMem models storage space of channel c
    cMem->setStorageSpaceChannel(c);
}

/**
 * createMappedChannelToConnection ()
 * The function models a channel mapped to a connection.
 */
void createMappedChannelToConnection(TimedSDFchannel *ch, Connection *cn)
{
    TimedSDFactor::Processor latency, tdma;
    SDFchannel *cConn, *cA, *cB, *cC, *cD, *cE;
    TimedSDFchannel *cSrc, *cDst;
    TimedSDFactor *c, *d, *e;
    Processor *dstP;
    SDFgraph *g;
    double connectionDelay;
    
    // SDF graph
    g = ch->getGraph();

    // Destination processor of connection
    dstP = cn->getDstTile()->getProcessor();
    
    // Create an actor which models the connection
    c = (TimedSDFactor*)g->createActor();
    c->setName(ch->getName() + "_connection");
    connectionDelay = ch->getTokenSize() / ch->getMinBandwidth();
    latency.execTime = cn->getLatency() + (CSize) (ceil(connectionDelay));
    latency.type = "latency";

    latency.stateSize = 0;
    c->addProcessor(&latency);
    c->setDefaultProcessor("latency");

    // Add self-loop to actor c (only 1 token sent at the same time)
    cConn = g->createChannel(c, 1, c, 1, 1);

    // Create channel to model source buffer
    cSrc = (TimedSDFchannel*) g->createChannel(c, 1, ch->getSrcActor(),
                                            ch->getSrcPort()->getRate(), 0);
    if ((CSize)ch->getBufferSize().src < ch->getInitialTokens())
        throw CException("Insufficient memory space for initial tokens.");
    cSrc->setInitialTokens(ch->getBufferSize().src - ch->getInitialTokens());

    // Channel cSrc models storage space of channel ch
    cSrc->setStorageSpaceChannel(ch);

    // Create channel to model destination buffer
    cDst = (TimedSDFchannel*)  g->createChannel(ch->getDstActor(),
                                            ch->getDstPort()->getRate(),c,1,0);
    cDst->setInitialTokens(ch->getBufferSize().dst);

    // Channel cSrc models storage space of channel ch
    cDst->setStorageSpaceChannel(ch);

    // Connect source actor channel ch to actor c
    cA = g->createChannel(ch->getSrcActor(),ch->getSrcPort()->getRate(),c, 1,0);
    cA->setInitialTokens(ch->getInitialTokens());

    // Synchronization between timewheels needed?
    if (dstP->getTimewheelSize() > dstP->getReservedTimeSlice())
    {
        // Create an actor which models the synchronization between TDMA wheels
        d = (TimedSDFactor*)g->createActor();
        d->setName(ch->getName() + "_tdma");
        tdma.type = "tdma";
        tdma.execTime = dstP->getTimewheelSize() - dstP->getReservedTimeSlice();
        tdma.stateSize = 0;
        d->addProcessor(&tdma);
        d->setDefaultProcessor("tdma");

        // The channel ch must be split in 3 channels connected to actor c and d
        cB = g->createChannel(c, 1, d, 1, 0);
        cC = g->createChannel(d, 1, ch->getDstActor(),
                                                ch->getDstPort()->getRate(), 0);
    }
    else
    {
        cC = g->createChannel(c, 1, ch->getDstActor(),
                                                ch->getDstPort()->getRate(), 0);
    }
    
    // Create actor to model minimal latency between production and
    // consumption of tokens on a channel bound to a connection.
    if (ch->getMinLatency() > 0)
    {
        e = (TimedSDFactor*)g->createActor();
        e->setName(ch->getName() + "_latency");
        latency.type = "latency";
        latency.execTime = ch->getMinLatency();
        latency.stateSize = 0;
        e->addProcessor(&latency);
        e->setDefaultProcessor("latency");

        cD = (TimedSDFchannel*) g->createChannel(ch->getSrcActor(),
                                        ch->getSrcPort()->getRate(), e, 1, 0);
        cE = (TimedSDFchannel*) g->createChannel(e, 1, ch->getDstActor(),
                                        ch->getDstPort()->getRate(), 0);
        cD->setInitialTokens(ch->getInitialTokens());
    }
}

/**
 * modelBindingAsTimedSDFgraph ()
 * Model all binding decisions in a timed SDF graph.
 */
TimedSDFgraph *modelBindingAsTimedSDFgraph(TimedSDFgraph *g, PlatformGraph *ag)
{
    Tile *t, *srcTile, *dstTile;
    Connections channelBinding;
    Tiles actorBinding;
    TimedSDFgraph *h;
    Connection *c;
    
    // Clone the SDF graph
    SDFcomponent component = SDFcomponent(g->getParent(), g->getId(), 
                                            g->getName());
    h = g->clone(component); 

    // Extract the binding of actors to tiles and channels to connections
    extractActorBinding(g, ag, actorBinding);
    extractChannelBinding(g, ag, channelBinding);

    // Set the execution time of each actor to the correct processor
    for (SDFactorsIter iter = g->actorsBegin(); iter != g->actorsEnd(); iter++)
    {
        SDFactor *a = *iter;
    
        // Tile
        t = actorBinding[a->getId()];
    
        // Actor not bound to a tile?
        if (t == NULL)
            throw CException("All actors must be bound to a tile.");

        createMappedActor((TimedSDFactor*)(h->getActor(a->getId())), t);
    }
    
    // Insert SDF model for each channel
    for (SDFchannelsIter iter = g->channelsBegin();
            iter != g->channelsEnd(); iter++)
    {
        TimedSDFchannel *ch = (TimedSDFchannel*)*iter;

        // Skip the channel if its a self-edge on an actor
        if (ch->getSrcActor()->getId() == ch->getDstActor()->getId())
            continue;
        
        // Tiles
        srcTile = actorBinding[ch->getSrcActor()->getId()];
        dstTile = actorBinding[ch->getDstActor()->getId()];
        
        // Source and destination actor bound to same tile?
        if (srcTile == dstTile)
        {
            // Channel should not be mapped to a connection
            if (channelBinding[ch->getId()] != NULL)
                throw CException("Channel should not be bound to connection.");
        
            createMappedChannelToTile((TimedSDFchannel*)
                                (h->getChannel(ch->getId())), srcTile);
        }
        else
        {
            c = channelBinding[ch->getId()];
             
            // Channel must be bound to a connection from source to destination
            // tile
            if (c == NULL || c->getSrcTile() != srcTile 
                    || c->getDstTile() != dstTile)
            {
                throw CException("Channel bound to invalid connection.");
            }
            
            createMappedChannelToConnection((TimedSDFchannel*)
                                (h->getChannel(ch->getId())), c);
        }
    }

    return h;
}
