/*
 *  TU Eindhoven
 *  Eindhoven, The Netherlands
 *
 *  Name            :   message.cc
 *
 *  Author          :   Sander Stuijk (sander@ics.ele.tue.nl)
 *
 *  Date            :   January 6, 2006
 *
 *  Function        :   Message
 *
 *  History         :
 *      06-01-06    :   Initial version.
 *
 * $Id: message.cc,v 1.1.1.1 2007/10/02 10:59:47 sander Exp $
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * 
 * In other words, you are welcome to use, share and improve this program.
 * You are forbidden to forbid anyone else to use, share and improve
 * what you give them.   Happy coding!
 */

#include "message.h"
#include "arch_graph.h"

/**
 * print ()
 * Output a message to the given stream.
 */
ostream &Message::print(ostream &out) const
{
    out << "(" << getId() << ", " << getSrcNodeId();
    out << ", " << getDstNodeId() << ", " << getStreamId();
    out << ", " << getSeqNr() << ", " << getStartTime() << ", ";
    out << getDuration() << ", " << getSize() << ")";
    
    return out;
}

/**
 * constructMessages ()
 * The function constructs a list of all messages from an XML document.
 */
void constructMessages(CNode *messagesNode, InterconnectGraph *archGraph,
        Messages &messages)
{
    Message *m, *prevMsg = NULL;
    Node *srcNode, *dstNode;
    CStrings channels;
    CString channel;
    CId streamId;
    CId nr;
    
    // Iterate over the list of messages
    for (CNode *messageNode = CGetChildNode(messagesNode, "message");
            messageNode != NULL; 
            messageNode = CNextNode(messageNode, "message"))
    {
        // Get id of message
        nr = CGetAttribute(messageNode, "nr");

        // Allocate memory for message
        m = new Message(nr);

        // Find source node in graph
        srcNode = archGraph->getNode(CGetAttribute(messageNode, "src"));
        if (srcNode == NULL)
            throw CException("[ERROR] Source node not found in architecture " \
                             "graph.");
        m->setSrcNodeId(srcNode->getId());

        // Find destination node in graph
        dstNode = archGraph->getNode(CGetAttribute(messageNode, "dst"));
        if (dstNode == NULL)
            throw CException("[ERROR] Destination node not found in " \
                             "architecture graph.");
        m->setDstNodeId(dstNode->getId());
        
        // Set stream id based on channel name
        channel = CGetAttribute(messageNode, "channel");
        streamId = 0;
        for (CStringsIter iter = channels.begin();
                iter != channels.end(); iter++)
        {
            if ((*iter) == channel)
                break;
            streamId++;
        }
        // No channel found with this name
        if (streamId == channels.size())
            channels.push_back(channel);
        m->setStreamId(streamId);
                
        // Set properties of message
        m->setSeqNr(CGetAttribute(messageNode, "seqNr"));
        m->setStartTime(CGetAttribute(messageNode, "startTime"));
        m->setDuration(CGetAttribute(messageNode, "duration"));
        m->setSize(CGetAttribute(messageNode, "size"));
        
        // Add message to the set of messages
        messages.push_back(m);
        
        // Create link between messages in stream
        if (prevMsg != NULL && prevMsg->getStreamId() == m->getStreamId())
        {
            // Check sequence order
            if (prevMsg->getSeqNr() >= m->getSeqNr())
                throw CException("Sequence order incorrect.");

            prevMsg->setNextMessageInStream(m);
            m->setPreviousMessageInStream(prevMsg);
        }
        prevMsg = m;
    }
}
