/*
 *  TU Eindhoven
 *  Eindhoven, The Netherlands
 *
 *  Name            :   sp_earliestdeadline.cc
 *
 *  Author          :   Sander Stuijk (sander@ics.ele.tue.nl)
 *
 *  Date            :   July 27, 2005
 *
 *  Function        :   Earliest Deadline First scheduler for single processor
 *
 *  History         :
 *      27-07-05    :   Initial version.
 *
 * $Id: sp_earliestdeadline.cc,v 1.1.1.1 2007/10/02 10:59:47 sander Exp $
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * 
 * In other words, you are welcome to use, share and improve this program.
 * You are forbidden to forbid anyone else to use, share and improve
 * what you give them.   Happy coding!
 */

#include "sp_earliestdeadline.h"
#include "../../transform/hsdf/hsdf.h"
#include "../../transform/to_apg/sdftoapg.h"

/**
 * SPearliestDeadlineScheduler ()
 * Constructor.
 */
SPearliestDeadlineScheduler::SPearliestDeadlineScheduler() : scheduler()
{
    schedule = NULL;
    hsdfGraph = NULL;
}
    
/**
 * ~SPearliestDeadlineScheduler ()
 * Desctructor
 */
SPearliestDeadlineScheduler::~SPearliestDeadlineScheduler()
{
}
    
/**
 * run ()
 * Run the scheduler on graph g.
 */
int SPearliestDeadlineScheduler::run(TimedSDFgraph *g)
{
    APGgraph *apgGraph;
    APGnodes readyList;
    
    // Convert SDF graph to HSDF graph
    hsdfGraph = (TimedSDFgraph*)transformSDFtoHSDF(g);

    // Create a schedule
    schedule = new SPschedule(hsdfGraph, g->getId());
    
    // Create acyclic precedence graph
    apgGraph = transformHSDFtoAPG(hsdfGraph, 1);

    // Compute deadlines
    computeDeadlines(apgGraph);

    // Compute initial ready list
    initReadyList(readyList, apgGraph);

    // Schedule all ready actors based on deadlines
    APGnode *n = getNodeWithEarliestDeadline(readyList);
    while (n != NULL)
    {
        // Schedule the actor corresponding to node n
        schedule->addActor(n->getId());

        // Update the ready list
        updateReadyList(readyList, n, apgGraph);

        // Get next node to be scheduled
        n = getNodeWithEarliestDeadline(readyList);
    }
    
    // Check that all nodes are scheduled
    ASSERT(apgGraph->nrNodes() == 0, "Not all nodes are scheduled.");
    
    // Cleanup
    delete apgGraph;
    
    // Done
    return SCHEDULER_OK;
}
    
/**
 * check ()
 * Check that all conditions are met to schedule the graph g.
 */
int SPearliestDeadlineScheduler::check(TimedSDFgraph *g)
{
    // No requirement, so conditions met
    return SCHEDULER_OK;
}
    
/**
 * reset ()
 * Reset the scheduler.
 */
void SPearliestDeadlineScheduler::reset()
{
    delete schedule;
    delete hsdfGraph;
    
    schedule = NULL;
    hsdfGraph = NULL;
}

/**
 * getNodeWithEarliestDeadline ()
 * Return a pointer to the node in the APG graph with the earliest
 * deadline (highest deadline value) from the list of ready nodes.
 * If the list is empty, a NULL pointer is returned.
 */
APGnode *SPearliestDeadlineScheduler::getNodeWithEarliestDeadline(
        APGnodes &readyList)
{
    APGnode *u = NULL;
    int maxWeight = -1;
    
    // Find node u in the ready list with the largest weight
    for (APGnodesIter iter = readyList.begin(); iter != readyList.end();
            iter++)
    {
        APGnode *n = *iter;
        
        if (n->getWeight() > maxWeight)
        {
            maxWeight = n->getWeight();
            u = n;
        }
    }
    
    return u;
}

/**
 * initReadyList ()
 * Initialize the node ready list for the APG graph g. Initially, all
 * nodes with no ingoing edges are ready. All edges are given a weight
 * of -1 to indicate that the connected nodes are not ready.
 */
void SPearliestDeadlineScheduler::initReadyList(APGnodes &readyList,
        APGgraph *g)
{
    // Clear ready list
    readyList.clear();
    
    // Add all nodes with indegree 0
    for (APGnodesIter iter = g->nodesBegin(); iter != g->nodesEnd(); iter++)
    {
        APGnode *n = *iter;
        
        if (n->inDegree() == 0)
            readyList.push_back(n);
    }
    
    // Set weight of all edges to -1
    for (APGedgesIter iter = g->edgesBegin(); iter != g->edgesEnd(); iter++)
    {
        APGedge *e = *iter;
        
        e->setWeight(-1);
    }
}

/**
 * updateReadyList ()
 * Update the node ready list by removing the actor v from it and adding
 * all nodes that become ready as the result of scheduling v. The weight of
 * all edges to which v is connected are updated.
 */
void SPearliestDeadlineScheduler::updateReadyList(APGnodes &readyList,
        APGnode *v, APGgraph *g)
{
    // Mark all edges outgoing from v as ready (set weight to zero)
    for (APGedgesIter iter = v->outEdgesBegin(); 
            iter != v->outEdgesEnd(); iter++)
    {
        APGedge *e = *iter;
        
        e->setWeight(0);
    }
    
    // Add all nodes adjacent to v for which all incoming edges have a
    // weight of zero
    for (APGedgesIter iter = v->outEdgesBegin(); 
            iter != v->outEdgesEnd(); iter++)
    {   
        bool ready = true;
        APGnode *u = v->opposite(*iter);
        
        // Check all incoming edges for non-infinite weight
        for (APGedgesIter iterU = u->inEdgesBegin(); iterU != u->inEdgesEnd();
                iterU++)
        {
            APGedge *e = *iterU;
            
            // Is adjacent actor not ready (negative weight on edge)?
            if (e->getWeight() < 0)
            {
                ready = false;
                break;
            }
        }
        
        // Node u not ready?
        if (!ready)
            continue;
        
        // Add node u to the ready list (if not yet in list)
        addNodeToReadyList(readyList, u);    
    }
    
    // Remove node v from the ready list
    for (APGnodesIter iter = readyList.begin();
            iter != readyList.end(); iter++)
    {
        APGnode *u = *iter;

        if (v->getId() == u->getId())
        {
            readyList.erase(iter);
            break;   
        }
    }
    
    // Remove node v from the APG graph
    g->delNode(v);
}

/**
 * computeDeadlines ()
 * The functions assign a weight to all nodes in the graph g. This weight
 * is equal to the deadline of the node. It is computed using a longest
 * path algorithm.
 */
void SPearliestDeadlineScheduler::computeDeadlines(APGgraph *g)
{
    APGnodes nodes;
    
    // Set weight of all edges to -1
    for (APGedgesIter iter = g->edgesBegin(); iter != g->edgesEnd(); iter++)
    {
        APGedge *e = *iter;
        
        e->setWeight(-1);
    }
    
    // Add all nodes with outdegree 0
    for (APGnodesIter iter = g->nodesBegin(); iter != g->nodesEnd(); iter++)
    {
        APGnode *n = *iter;
        
        if (n->outDegree() == 0)
            nodes.push_back(n);
    }
    
    while (!nodes.empty())
    {
        APGnodes nextNodes;
        
        // Set weight of all edges connected to the nodes and find next nodes
        // (i.e. nodes of which all outgoing edges get a non-negative weight
        // because of a weight assignment in this step).
        for (APGnodesIter iter = nodes.begin(); iter != nodes.end(); iter++)
        {
            APGnode *n = *iter;
            
            // Set the weight of the node
            n->setWeight(n->getWeight() + maxWeightOnAllOutEdges(n));

            // Set weight of all edges incident to the node n
            for (APGedgesIter iterE = n->inEdgesBegin();
                    iterE != n->inEdgesEnd(); iterE++)
            {
                APGedge *e = *iterE;
                APGnode *m = e->opposite(n);
                
                e->setWeight(n->getWeight());
                
                // Check that opposite node has non-negative weight on all of 
                // its outgoing edges
                if (positiveWeightOnAllOutEdges(m))
                {
                    // Add node to list of next nodes
                    nextNodes.push_back(m);
                }
                
            }
        }
        
        nodes = nextNodes;
    }
}

/**
 * maxWeightOnAllOutEdges ()
 * The function returns the maximum weight found on all outgoing edges of
 * node n. If a node has no outgoing edges, the maximum is defined to be zero.
 */
int SPearliestDeadlineScheduler::maxWeightOnAllOutEdges(APGnode *n)
{
    int maxWeight = INT_MIN;
    
    // No outgoing edges?
    if (n->outEdgesBegin() == n->outEdgesEnd())
        return 0;
    
    // Find max weight
    for (APGedgesIter iter = n->outEdgesBegin();
            iter != n->outEdgesEnd(); iter++)
    {
        APGedge *e = *iter;
        
        if (e->getWeight() > maxWeight)
            maxWeight = e->getWeight();
    }
    
    return maxWeight;
}

/**
 * positiveWeightOnAllOutEdges ()
 * The function returns true if all of outgoing edges of node n have
 * a non-negative weight. Else it returns false.
 */
bool SPearliestDeadlineScheduler::positiveWeightOnAllOutEdges(APGnode *n)
{
    for (APGedgesIter iter = n->outEdgesBegin();
            iter != n->outEdgesEnd(); iter++)
    {
        APGedge *e = *iter;
        
        if (e->getWeight() < 0)
            return false;
    }
    
    return true;
}

/**
 * addNodeToReadyList ()
 * The function add the node n to the readyList if it is not yet present in
 * the list.
 */
void SPearliestDeadlineScheduler::addNodeToReadyList(APGnodes &readyList,
        APGnode *n)
{
    // Check that node is not already in list
    for (APGnodesIter iter = readyList.begin(); iter != readyList.end(); iter++)
    {
        APGnode *m = *iter;
        
        if (m->getId() == n->getId())
            return;
    }
    
    // Add node to list
    readyList.push_back(n);
}
