/*
 *  TU Eindhoven
 *  Eindhoven, The Netherlands
 *
 *  Name            :   binding.cc
 *
 *  Author          :   Sander Stuijk (sander@ics.ele.tue.nl)
 *
 *  Date            :   April 11, 2006
 *
 *  Function        :   Base class for all binding algorithms
 *
 *  History         :
 *      11-04-06    :   Initial version.
 *
 * $Id: binding.cc,v 1.3 2008/01/24 12:43:56 sander Exp $
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * 
 * In other words, you are welcome to use, share and improve this program.
 * You are forbidden to forbid anyone else to use, share and improve
 * what you give them.   Happy coding!
 */

#include "binding.h"
#include "../mpsoc_arch/sdf.h"
#include "../../transform/model/autoconc.h"
#include "../../transform/model/buffersize.h"
#include "../../analysis/statespace/statespace.h"
#include "../../output/xml/xml.h"

/**
 * analyzeThroughputApplication ()
 * The function returns the throughput of the application graph
 * considering the absense of auto-concurrency, the buffer constraints
 * and the mapping of actors to the fastest processor.
 */
double Binding::analyzeThroughputApplication()
{
    SDFstateSpaceThroughputAnalysis thrAnalysisAlgo;
    TimedSDFgraph *g, *gr;
    double thrGraph;

    // Model absense of auto-concurrency in SDFG
    g = (TimedSDFgraph*)modelAutoConcurrencyInSDFgraph(appGraph, 1);
    
    // Set buffer size of each channel to max(mem, src+dst)
    for (SDFchannelsIter iter = g->channelsBegin();
            iter != g->channelsEnd(); iter++)
    {
        TimedSDFchannel *c = (TimedSDFchannel*)*iter;
        TimedSDFchannel::BufferSize sz = c->getBufferSize();
        
        if (sz.mem < sz.src + sz.dst)
        {
            sz.sz = sz.src+sz.dst;
            c->setBufferSize(sz);
        }    
    }
    
    // Model buffer constraints in SDFG
    gr = modelBufferSizeInSDFgraph(g);

    // Select fastest processor for each actor as the default processor
    for (SDFactorsIter iter = gr->actorsBegin(); 
            iter != gr->actorsEnd(); iter++)
    {
        TimedSDFactor *a = (TimedSDFactor*)*iter;
        TimedSDFactor::Processor *proc = NULL;
        
        for (TimedSDFactor::ProcessorsIter iterP = a->processorsBegin();
                iterP != a->processorsEnd(); iterP++)
        {
            TimedSDFactor::Processor *p = *iterP;
            
            if (proc == NULL || proc->execTime > p->execTime)
                proc = p;
        }
        
        if (proc != NULL)
            a->setDefaultProcessor(proc->type);
    }

    // Compute throughput of the SDFG
    thrGraph = thrAnalysisAlgo.analyze(gr);

    // Cleanup
    delete g;
    delete gr;
    
    return thrGraph;
}

/**
 * analyzeThroughput ()
 * The function returns the throughput of the application graph
 * mapped onto the platform graph.
 */
double Binding::analyzeThroughput(vector<double> &tileUtilization)
{
    SDFstateSpaceBindingAwareThroughputAnalysis thrAnalysisAlgo;
    TimedSDFgraph *mappedAppGraph;
    double thrGraph;

    // Model mapping in application SDF graph
    mappedAppGraph = modelBindingAsTimedSDFgraph(appGraph, archGraph);

    // Compute throughput of mapped SDF graph
    thrGraph = thrAnalysisAlgo.analyze(mappedAppGraph, archGraph,
                                            tileUtilization);

    // Cleanup
    delete mappedAppGraph;

    return thrGraph;
}

/**
 * isThroughputConstraintSatisfied ()
 * Check wether or not the throughput constraint is satisfied.
 */
bool Binding::isThroughputConstraintSatisfied()
{
    double thrGraph, thrConstraint;
    vector<double> tileUtilization;
    
    // Compute throughput of mapped SDF graph
    thrGraph = analyzeThroughput(tileUtilization);
    thrConstraint = appGraph->getThroughputConstraint().value();

    #ifdef VERBOSE
    cerr << "Throughput: " << thrGraph << endl;
    cerr << "Constraint: " << thrConstraint << endl;
    #endif

    // Is throughput constraint met?
    if (thrConstraint > thrGraph)
    {
        #ifdef VERBOSE
        cerr << "Failed throughput constraint." << endl;
        #endif
        
        return false;
    }

    #ifdef VERBOSE
    cerr << "Throughput constraint satisfied." << endl;
    #endif
    
    return true;
}

/**
 * sizeOfStaticOrderSchedule ()
 * The function returns the number of states in the schedule.
 */
uint Binding::sizeOfStaticOrderSchedule(Schedule *s)
{
    uint nrStates = 0;
    
    while (s != NULL)
    {
        nrStates++;
        
        // Reached last state?
        if (s->last)
            break;
        
        // Next state
        s = s->next;
    }
    
    return nrStates;
}

/**
 * minimizeStaticOrderSchedule ()
 * The function reduces the length of a schedule s by finding the smallest
 * repeated part in the periodic phase of the schedule. The periodic phase
 * of the minimized schedule contains this repeated part once. From the
 * transient phase, all complete repetitions of the periodic phase are
 * removed. The function returns a pointer to the minimized schedule. Note
 * that the schedule 's' is lost. Schedule elements which are not part of the
 * minimized schedule are freed and the schedule elements are assigned a new id.
 */
Schedule *Binding::minimizeStaticOrderSchedule(Schedule *s)
{
    Schedule *periodicStart, *periodicEnd, *transientEnd;
    Schedule *posPattern, *posSchedule;
    bool startOfPattern = true;
    Schedule *x, *y;
    uint id;

    // Empty schedule?
    if (s == NULL)
        return s;

    // Find start of periodic schedule
    periodicStart = s;
    while (periodicStart != NULL && !periodicStart->last) 
    { 
        periodicStart = periodicStart->next; 
    }
    ASSERT(periodicStart, "periodicStart == NULL: no periodic regime.");
    periodicStart = periodicStart->next;
    
    // Find smallest repeated part in periodic schedule
    periodicEnd = periodicStart;
    posPattern = periodicStart;
    posSchedule = periodicEnd->next;

    while (posSchedule != periodicStart)
    {
        // Examining position in schedule (so no longer at start of pattern)
        startOfPattern = false;
        
        // Actor at posSchedule not equal to actor at posPattern?
        if (posSchedule->actor->getId() != posPattern->actor->getId())
        {
            // Add schedule part up-to and including posSchedule to the
            // pattern and restart the search starting from the next
            // position in the schedule
            periodicEnd = posSchedule;
            posPattern = periodicStart;
        }
        else
        {
            // Actor in schedule matches with actor in pattern, move one
            // position forward in pattern and in schedule (loop on pattern
            // if needed).
            posPattern = posPattern->next;
            if (posPattern->id == periodicEnd->next->id)
            {
                posPattern = periodicStart;
                startOfPattern = true;
            }
        }
        
        // Last element reached?
        if (posSchedule->last)
            break;
        
        // Next
        posSchedule = posSchedule->next;
    }

    // Is the exact end of the repeated part of the schedule not reached?
    if (!startOfPattern)
    {
        // No exact repetition of pattern
        periodicEnd = posSchedule;
    }

    // Remove actor from non-repeated part of the schedule
    // An part can be removed if (starting from the back), the
    // complete repeated pattern is found in the transient part.
    // If not, no further reduction of the transient part is
    // possible.
    transientEnd = periodicStart->previous;
    posPattern = periodicEnd;
    posSchedule = transientEnd;
    while (posSchedule != NULL)
    {
        if (posSchedule->actor->getId() != posPattern->actor->getId())
        {
            break;
        }
        else
        {
            posSchedule = posSchedule->previous;
            if (posPattern->id == periodicStart->id)
            {
                posPattern = periodicEnd;
                transientEnd = posSchedule;
            }
            else
            {
                posPattern = posPattern->previous;
            }
        }
    }

    // Special case: the transient phase contains no pattern which does not
    // match with the pattern in the periodic phase. In this case, the schedule
    // is periodic from the start onwards. Note that it is possible that the
    // transient phase starts with an incomplete part of the repeated pattern.
    //
    // Example: 
    //      Schedule: a c d a c d a c
    //                    ^         |
    //                    +---------+
    //
    //      This schedule is has as a pattern 'dac'. The transient part contains
    //      an incomplete part of the pattern. This schedule is periodic from
    //      the start with the pattern 'acd'.
    if (posSchedule == NULL && transientEnd != NULL)
    {
        // Last element of the pattern not found in the search becomes last
        // element of the pattern and loops back to the start
        posPattern->last = true;
        posPattern->next = s;
        
        // Done
        return s;
    }

    // Cleanup unused schedule elements in periodic phase
    x = periodicEnd->next;
    while (x != periodicStart)
    {
        // Last element reached?
        if (x->last)
        {
            delete x;
            break;
        }

        // Pointer to next element in the schedule
        y = x->next;

        // Cleanup
        delete x;

        // Next
        x = y;
    }

    // Remove repeated pattern from the periodic part
    periodicEnd->last = true;
    periodicEnd->next = periodicStart;

    // Cleanup unused schedule elements in transient phase
    if (transientEnd == NULL)
        x = s;
    else
        x = transientEnd->next;
    while (x != periodicStart)
    {
        y = x->next;
        delete x;
        x = y;
    }
    
    // Remove repeated pattern from the transient phase
    periodicStart->previous = transientEnd;
    if (transientEnd != NULL)
        transientEnd->next = periodicStart;

    // Schedule starts in the periodic phase
    if (transientEnd == NULL)
        s = periodicStart;

    // Renumber states in the schedule
    id = 0;
    posSchedule = s;
    while (posSchedule != NULL)
    {
        posSchedule->id = id;
    
        // Last element reached?
        if (posSchedule->last)
            break;
    
        // Next
        posSchedule = posSchedule->next;
        id++;
    }    

    // Done
    return s;   
}

/**
 * minimizeStaticOrderSchedules ()
 * The function minimizes the length of all static-order schedules.
 */
void Binding::minimizeStaticOrderSchedules(PlatformGraph *archGraph)
{
    // Iterate over the tiles
    for (TilesIter iter = archGraph->tilesBegin();
            iter != archGraph->tilesEnd(); iter++)
    {
        Tile *t = *iter;
        Processor *p = t->getProcessor();
        
        // Processor inside the tile
        if (p != NULL)
        {
            Schedule *s = p->getSchedule();
            
            // Schedule given for the processor
            if (s != NULL)
            {
                s = minimizeStaticOrderSchedule(s);
                p->setSchedule(s);
                
                logMsg("Reduced schedule on '" + t->getName() + "' to "
                       + CString(sizeOfStaticOrderSchedule(s)) + " states.");
            }
        }
    }
}
